# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'hardware_tests.ui'
##
## Created by: Qt User Interface Compiler version 5.15.2
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide2.QtCore import *
from PySide2.QtGui import *
from PySide2.QtWidgets import *


class Ui_HardwareTestDialog(object):
    def setupUi(self, HardwareTestDialog):
        if not HardwareTestDialog.objectName():
            HardwareTestDialog.setObjectName(u"HardwareTestDialog")
        HardwareTestDialog.resize(754, 426)
        self.verticalLayout = QVBoxLayout(HardwareTestDialog)
        self.verticalLayout.setObjectName(u"verticalLayout")
        self.testOutput = QPlainTextEdit(HardwareTestDialog)
        self.testOutput.setObjectName(u"testOutput")
        self.testOutput.setVerticalScrollBarPolicy(Qt.ScrollBarAlwaysOn)
        self.testOutput.setReadOnly(True)

        self.verticalLayout.addWidget(self.testOutput)

        self.buttonBox = QDialogButtonBox(HardwareTestDialog)
        self.buttonBox.setObjectName(u"buttonBox")
        self.buttonBox.setOrientation(Qt.Horizontal)
        self.buttonBox.setStandardButtons(QDialogButtonBox.Close|QDialogButtonBox.Reset)

        self.verticalLayout.addWidget(self.buttonBox)


        self.retranslateUi(HardwareTestDialog)
        self.buttonBox.accepted.connect(HardwareTestDialog.accept)
        self.buttonBox.rejected.connect(HardwareTestDialog.reject)

        QMetaObject.connectSlotsByName(HardwareTestDialog)
    # setupUi

    def retranslateUi(self, HardwareTestDialog):
        HardwareTestDialog.setWindowTitle(QCoreApplication.translate("HardwareTestDialog", u"Hardware Tests", None))
    # retranslateUi




import logging

from PySide2 import QtCore, QtGui, QtWidgets

from hyperborea.preferences import read_bool_setting

from .ui_tcp_scan_dialog import Ui_TCPScanDialog

logger = logging.getLogger(__name__)


class TCPScanDialog(QtWidgets.QDialog, Ui_TCPScanDialog):
    def __init__(self, devices, get_location_strings, rescan_func,
                 parent=None):
        super().__init__(parent)

        self.get_location_strings = get_location_strings
        self.rescan_func = rescan_func

        self.settings = QtCore.QSettings()

        self.device_by_row = []

        self.setupUi(self)
        self.extra_ui_setup()

        self.tableWidget.itemSelectionChanged.connect(self.selection_changed)

        self.update_devices(devices)
        self.selection_changed()

    def extra_ui_setup(self):
        header = self.tableWidget.horizontalHeader()
        header.setResizeMode(QtWidgets.QHeaderView.ResizeToContents)

        self.rescanButton = self.buttonBox.button(
            QtWidgets.QDialogButtonBox.Reset)
        self.rescanButton.setText(self.tr("Rescan"))
        self.rescanButton.clicked.connect(self.rescan)

        self.tableWidget.cellDoubleClicked.connect(self.double_click_cb)
        self.tableWidget.itemDoubleClicked.connect(self.double_click_cb)

        self.rescan_timer = QtCore.QTimer(self)
        self.rescan_timer.timeout.connect(self.rescan)

        self.automaticRescan.toggled.connect(self.set_automatic_rescan)

        automatic_rescan = read_bool_setting(self.settings, "AutomaticRescan",
                                             True)
        self.automaticRescan.setChecked(automatic_rescan)

        self.finished.connect(self.rescan_timer.stop)

    def set_automatic_rescan(self, _junk=None):
        if self.automaticRescan.isChecked():
            self.settings.setValue("AutomaticRescan", 1)
            self.rescan_timer.start(1000)
        else:
            self.settings.setValue("AutomaticRescan", 0)
            self.rescan_timer.stop()

    def double_click_cb(self, _row=None, _column=None):
        self.accept()

    def sort_devices(self, all_devices):
        def key_func(t):
            device, _connected_bool = t
            adv = device.tcp_get_advertisement()
            return (adv.connected, adv.serial_number)
        return sorted(all_devices, key=key_func)

    def update_device_row(self, device, connected, row_index):
        adv = device.tcp_get_advertisement()
        is_boot = device.supports_bootloader_commands()

        if connected:
            item_flags = QtCore.Qt.NoItemFlags
            available_str = "No (tab already open)"
        elif adv.connected:
            item_flags = QtCore.Qt.NoItemFlags
            available_str = "No (in use)"
        else:
            item_flags = QtCore.Qt.ItemIsSelectable | QtCore.Qt.ItemIsEnabled
            available_str = "Yes"

        serial_number_item = QtWidgets.QTableWidgetItem(adv.serial_number)
        serial_number_item.setFlags(item_flags)
        self.tableWidget.setItem(row_index, 0, serial_number_item)

        user_tag1_item = QtWidgets.QTableWidgetItem(adv.user_tag1)
        user_tag1_item.setFlags(item_flags)
        self.tableWidget.setItem(row_index, 1, user_tag1_item)

        user_tag2_item = QtWidgets.QTableWidgetItem(adv.user_tag2)
        user_tag2_item.setFlags(item_flags)
        self.tableWidget.setItem(row_index, 2, user_tag2_item)

        board_info_str = "{} rev {}".format(adv.board_type, adv.board_rev)
        board_info_item = QtWidgets.QTableWidgetItem(board_info_str)
        board_info_item.setFlags(item_flags)
        self.tableWidget.setItem(row_index, 3, board_info_item)

        build_info_item = QtWidgets.QTableWidgetItem(adv.build_info)
        build_info_item.setFlags(item_flags)
        self.tableWidget.setItem(row_index, 4, build_info_item)

        build_date_item = QtWidgets.QTableWidgetItem(adv.build_date)
        build_date_item.setFlags(item_flags)
        self.tableWidget.setItem(row_index, 5, build_date_item)

        if is_boot:
            bootloader_item = QtWidgets.QTableWidgetItem("Running")
            font = QtGui.QFont()
            font.setBold(True)
            bootloader_item.setFont(font)
        else:
            bootloader_item = QtWidgets.QTableWidgetItem("")
        bootloader_item.setFlags(item_flags)
        self.tableWidget.setItem(row_index, 6, bootloader_item)

        available_item = QtWidgets.QTableWidgetItem(available_str)
        available_item.setFlags(item_flags)
        self.tableWidget.setItem(row_index, 7, available_item)

        # highlight the whole row yellow if it's a bootloader
        if is_boot:
            yellow_brush = QtGui.QBrush(QtGui.QColor(QtCore.Qt.yellow))
            serial_number_item.setBackground(yellow_brush)
            user_tag1_item.setBackground(yellow_brush)
            user_tag2_item.setBackground(yellow_brush)
            board_info_item.setBackground(yellow_brush)
            build_info_item.setBackground(yellow_brush)
            build_date_item.setBackground(yellow_brush)
            bootloader_item.setBackground(yellow_brush)
            available_item.setBackground(yellow_brush)

            black_brush = QtGui.QBrush(QtGui.QColor(QtCore.Qt.black))
            serial_number_item.setForeground(black_brush)
            user_tag1_item.setForeground(black_brush)
            user_tag2_item.setForeground(black_brush)
            board_info_item.setForeground(black_brush)
            build_info_item.setForeground(black_brush)
            build_date_item.setForeground(black_brush)
            bootloader_item.setForeground(black_brush)
            available_item.setForeground(black_brush)

    def update_devices(self, devices):
        connected_location_strs = self.get_location_strings()
        all_devices = []
        for device in devices:
            location_str = device.get_location_string()
            connected_bool = location_str in connected_location_strs
            all_devices.append((device, connected_bool))
        all_devices = self.sort_devices(all_devices)

        # clear the existing devices
        self.tableWidget.setRowCount(len(all_devices))
        self.device_by_row.clear()

        for row_index, (device, connected_bool) in enumerate(all_devices):
            self.device_by_row.append(device)
            self.update_device_row(device, connected_bool, row_index)

    def get_selected_devices(self):
        selected_devices = []
        for index in self.tableWidget.selectionModel().selectedRows():
            selected_devices.append(self.device_by_row[index.row()])
        return selected_devices

    def get_selected_serials(self):
        selected_devices = self.get_selected_devices()
        serials = []
        for device in selected_devices:
            adv = device.tcp_get_advertisement()
            serials.append(adv.serial_number)
        return serials

    def select_serials(self, serials):
        self.tableWidget.setSelectionMode(
            QtWidgets.QAbstractItemView.MultiSelection)

        self.tableWidget.clearSelection()
        for i, device in enumerate(self.device_by_row):
            adv = device.tcp_get_advertisement()
            serial = adv.serial_number
            if serial in serials:
                self.tableWidget.selectRow(i)

        self.tableWidget.setSelectionMode(
            QtWidgets.QAbstractItemView.ExtendedSelection)

        self.selection_changed()

    def selection_changed(self):
        ok_button = self.buttonBox.button(QtWidgets.QDialogButtonBox.Ok)
        if self.tableWidget.selectionModel().hasSelection():
            ok_button.setEnabled(True)
        else:
            ok_button.setEnabled(False)

    def rescan(self):
        self.rescanButton.setEnabled(False)

        selected_serials = self.get_selected_serials()

        new_devices = self.rescan_func()
        self.update_devices(new_devices)

        if self.get_selected_serials() != selected_serials:
            self.select_serials(selected_serials)

        self.rescanButton.setEnabled(True)



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'ctrl_var_panel.ui'
##
## Created by: Qt User Interface Compiler version 5.15.2
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide2.QtCore import *
from PySide2.QtGui import *
from PySide2.QtWidgets import *


class Ui_CtrlVarPanel(object):
    def setupUi(self, CtrlVarPanel):
        if not CtrlVarPanel.objectName():
            CtrlVarPanel.setObjectName(u"CtrlVarPanel")
        CtrlVarPanel.resize(99, 41)
        self.verticalLayout = QVBoxLayout(CtrlVarPanel)
        self.verticalLayout.setObjectName(u"verticalLayout")
        self.ctrlVarLayout = QVBoxLayout()
        self.ctrlVarLayout.setObjectName(u"ctrlVarLayout")

        self.verticalLayout.addLayout(self.ctrlVarLayout)

        self.verticalSpacer = QSpacerItem(10, 10, QSizePolicy.Minimum, QSizePolicy.Expanding)

        self.verticalLayout.addItem(self.verticalSpacer)


        self.retranslateUi(CtrlVarPanel)

        QMetaObject.connectSlotsByName(CtrlVarPanel)
    # setupUi

    def retranslateUi(self, CtrlVarPanel):
        CtrlVarPanel.setWindowTitle(QCoreApplication.translate("CtrlVarPanel", u"Control Variable Panel", None))
        CtrlVarPanel.setTitle(QCoreApplication.translate("CtrlVarPanel", u"Control Variables", None))
    # retranslateUi




# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'info_dialog.ui'
##
## Created by: Qt User Interface Compiler version 5.15.2
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide2.QtCore import *
from PySide2.QtGui import *
from PySide2.QtWidgets import *


class Ui_InfoDialog(object):
    def setupUi(self, InfoDialog):
        if not InfoDialog.objectName():
            InfoDialog.setObjectName(u"InfoDialog")
        InfoDialog.resize(736, 548)
        self.verticalLayout = QVBoxLayout(InfoDialog)
        self.verticalLayout.setObjectName(u"verticalLayout")
        self.plainTextEdit = QPlainTextEdit(InfoDialog)
        self.plainTextEdit.setObjectName(u"plainTextEdit")
        self.plainTextEdit.setVerticalScrollBarPolicy(Qt.ScrollBarAlwaysOn)
        self.plainTextEdit.setReadOnly(True)

        self.verticalLayout.addWidget(self.plainTextEdit)

        self.buttonBox = QDialogButtonBox(InfoDialog)
        self.buttonBox.setObjectName(u"buttonBox")
        self.buttonBox.setOrientation(Qt.Horizontal)
        self.buttonBox.setStandardButtons(QDialogButtonBox.Close)

        self.verticalLayout.addWidget(self.buttonBox)


        self.retranslateUi(InfoDialog)
        self.buttonBox.accepted.connect(InfoDialog.accept)
        self.buttonBox.rejected.connect(InfoDialog.reject)

        QMetaObject.connectSlotsByName(InfoDialog)
    # setupUi

    def retranslateUi(self, InfoDialog):
        InfoDialog.setWindowTitle(QCoreApplication.translate("InfoDialog", u"Device Information", None))
    # retranslateUi




#!/usr/bin/env python3
import logging.handlers
import multiprocessing
import os
import sys
import time

from PySide2 import QtCore, QtGui, QtWidgets

import acheron
import asphodel
import hyperborea.proxy

from .plotmain import PlotMainWindow
from . import acheron_rc  # needed for the icon

logger = logging.getLogger(__name__)


def main_is_frozen():
    """Return True if the script is frozen, False otherwise."""
    return getattr(sys, 'frozen', False)


def setup_logging():
    """Configure logging for the whole program."""

    def my_excepthook(exctype, value, traceback):
        """Log the caught exception using the logging module."""
        exc_info = (exctype, value, traceback)
        logger.error("Uncaught Exception", exc_info=exc_info)

    sys.excepthook = my_excepthook

    # get an appropriate location for the log file
    logdir = QtCore.QStandardPaths.writableLocation(
        QtCore.QStandardPaths.DataLocation)
    logfile = os.path.join(logdir, "main.log")

    # make sure the directory exists
    os.makedirs(logdir, exist_ok=True)

    # filter for creating %(optdevice) format option
    optdevice_filter = hyperborea.proxy.OptionalDeviceStringFilter("[%s] ", "")

    # create a log file handler
    file_log_handler = logging.handlers.RotatingFileHandler(
        logfile, maxBytes=10e6, backupCount=1, encoding='utf-8')
    file_formatter = logging.Formatter(
        "%(asctime)s - %(name)s - %(levelname)s - %(optdevice)s%(message)s")
    file_formatter.default_time_format = '%Y-%m-%dT%H:%M:%S'
    file_formatter.default_msec_format = '%s,%03dZ'
    file_formatter.converter = time.gmtime
    file_log_handler.addFilter(optdevice_filter)
    file_log_handler.setFormatter(file_formatter)
    file_log_handler.setLevel(logging.DEBUG)

    root_logger = logging.getLogger()
    root_logger.addHandler(file_log_handler)
    root_logger.setLevel(logging.DEBUG)

    # remove pyusb's logging info
    pyusb_logger = logging.getLogger("usb")
    pyusb_logger.propagate = False

    # suppress boto3 (aws) debug messages
    logging.getLogger('boto3').setLevel(logging.INFO)
    logging.getLogger('botocore').setLevel(logging.INFO)
    logging.getLogger('s3transfer').setLevel(logging.INFO)
    logging.getLogger('urllib3').setLevel(logging.INFO)


def force_exit():
    # based on pyqtgraph.exit, with fix for macos

    # # invoke atexit callbacks
    import atexit
    atexit._run_exitfuncs()

    # # close file handles
    if sys.platform == 'darwin':
        # trying to close 7 produces an illegal instruction on the Mac
        os.closerange(3, 7)
        os.closerange(8, 4096)
    else:
        os.closerange(3, 4096)  # just guessing on the maximum descriptor count

    os._exit(0)


def main():
    """Run the program."""

    # freeze_support() MUST be first. Anything before this will cease to exist
    multiprocessing.freeze_support()

    # spawn is only option on windows, linux needs spawn or forkserver to work
    multiprocessing.set_start_method("spawn")

    # work around for botocore data
    if main_is_frozen():
        cacert = os.path.join(os.path.dirname(sys.executable), 'botodata',
                              'cacert.pem')
        os.environ["AWS_CA_BUNDLE"] = cacert
        botodata = os.path.join(os.path.dirname(sys.executable), 'botodata')
        os.environ["AWS_DATA_PATH"] = botodata

    if not sys.stdout or not sys.stderr:
        sys.stdout = open(os.devnull)
        sys.stderr = open(os.devnull)

    QtWidgets.QApplication.setAttribute(QtCore.Qt.AA_EnableHighDpiScaling)
    QtWidgets.QApplication.setAttribute(QtCore.Qt.AA_UseHighDpiPixmaps)

    app = QtWidgets.QApplication(sys.argv)
    app.setApplicationName("Acheron")
    app.setOrganizationDomain("suprocktech.com")
    app.setOrganizationName("Suprock Tech")

    QtGui.QIcon.setThemeName("suprock")

    # load the application icon
    icon = QtGui.QIcon()
    icon_reader = QtGui.QImageReader(":/acheron.ico")
    while True:
        pixmap = QtGui.QPixmap.fromImage(icon_reader.read())
        icon.addPixmap(pixmap)
        if not icon_reader.jumpToNextImage():
            break
    app.setWindowIcon(icon)

    # remove "What's This" button across the whole application
    app.setAttribute(QtCore.Qt.AA_DisableWindowContextHelpButton)

    # force the settings to use an INI file instead of the registry
    QtCore.QSettings.setDefaultFormat(QtCore.QSettings.IniFormat)

    app.setApplicationVersion(acheron.__version__)

    # add selectable text to all message boxes
    app.setStyleSheet("QMessageBox { messagebox-text-interaction-flags: 5; }")

    setup_logging()

    logger.info("Acheron started (Version {})".format(acheron.__version__))

    try:
        missing_funcs = asphodel.nativelib.missing_funcs
    except AttributeError:
        missing_funcs = []  # to bypass the next warning
        message = "Asphodel python mismatch!"
        logging.warning(message)
        QtWidgets.QMessageBox.warning(None, "Warning", message)

    if missing_funcs:
        missing_str = ", ".join(sorted(missing_funcs))
        logging.warning("Missing Asphodel functions: {}".format(missing_str))
        QtWidgets.QMessageBox.warning(None, "Warning",
                                      "Asphodel library mismatch!")

    proxy_manager = hyperborea.proxy.DeviceProxyManager()

    mainwin = PlotMainWindow(proxy_manager)
    mainwin.show()
    app.exec_()

    logger.info("Acheron main window closed")

    proxy_manager.stop()
    if mainwin.upload_manager:
        mainwin.upload_manager.stop()
        mainwin.upload_manager.join()

    logger.info("Acheron finished")

    force_exit()


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'led_control_widget.ui'
##
## Created by: Qt User Interface Compiler version 5.15.2
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide2.QtCore import *
from PySide2.QtGui import *
from PySide2.QtWidgets import *


class Ui_LEDControlWidget(object):
    def setupUi(self, LEDControlWidget):
        if not LEDControlWidget.objectName():
            LEDControlWidget.setObjectName(u"LEDControlWidget")
        LEDControlWidget.resize(176, 20)
        self.horizontalLayout_3 = QHBoxLayout(LEDControlWidget)
        self.horizontalLayout_3.setContentsMargins(0, 0, 0, 0)
        self.horizontalLayout_3.setObjectName(u"horizontalLayout_3")
        self.slider = QSlider(LEDControlWidget)
        self.slider.setObjectName(u"slider")
        self.slider.setMinimumSize(QSize(128, 0))
        self.slider.setMaximum(255)
        self.slider.setOrientation(Qt.Horizontal)

        self.horizontalLayout_3.addWidget(self.slider)

        self.spinBox = QSpinBox(LEDControlWidget)
        self.spinBox.setObjectName(u"spinBox")
        self.spinBox.setMaximum(255)

        self.horizontalLayout_3.addWidget(self.spinBox)

        QWidget.setTabOrder(self.slider, self.spinBox)

        self.retranslateUi(LEDControlWidget)
        self.slider.valueChanged.connect(self.spinBox.setValue)
        self.spinBox.valueChanged.connect(self.slider.setValue)

        QMetaObject.connectSlotsByName(LEDControlWidget)
    # setupUi

    def retranslateUi(self, LEDControlWidget):
        LEDControlWidget.setWindowTitle(QCoreApplication.translate("LEDControlWidget", u"LED Control Widget", None))
    # retranslateUi




import bisect
import collections
import datetime
import logging
import time
import threading

from PySide2 import QtCore, QtGui, QtWidgets

import asphodel
from hyperborea.preferences import read_bool_setting
import hyperborea.device_info
import hyperborea.proxy

from .ui_radio_scan_dialog import Ui_RadioScanDialog

logger = logging.getLogger(__name__)


instances = {}


ScanResult = collections.namedtuple(
    "ScanResult", ["active_scan", "serial_number", "is_bootloader",
                   "device_mode", "scan_strength", "last_seen", "user_tag_1",
                   "user_tag_2", 'board_info', 'build_info', 'build_date',
                   'has_bootloader'])


def start_radio_scan_manager(device, *args, **kwargs):
    device_logger = hyperborea.proxy.get_device_logger(logger, device)
    device_logger.info("Starting radio scan manager")

    if device in instances:
        stop_radio_scan_manager(device)

    instance = RadioScanManager(device, device_logger, *args, **kwargs)
    instances[device] = instance
    instance.start()

    hyperborea.proxy.register_local_cleanup(stop_radio_scan_manager, device)


def stop_radio_scan_manager(device):
    if device not in instances:
        return

    instances[device].stop()

    device_logger = hyperborea.proxy.get_device_logger(logger, device)
    device_logger.info("Stopped radio scan manager")

    del instances[device]

    hyperborea.proxy.unregister_local_cleanup(stop_radio_scan_manager, device)


class RadioScanManager:
    def __init__(self, device, logger, diskcache, status_pipe, control_pipe):
        self.device = device
        self.remote = device.get_remote_device()
        self.logger = logger
        self.diskcache = diskcache
        self.status_pipe = status_pipe
        self.control_pipe = control_pipe

        self.supports_scan_power = True
        self.power_max_queries = min(
            self.device.get_max_outgoing_param_length() // 4,
            self.device.get_max_incoming_param_length())

        self.finished = threading.Event()
        self.scan_thread = threading.Thread(target=self.scan_thread_run)

    def start(self):
        self.scan_thread.start()
        self.remote.open()

    def stop(self):
        self.finished.set()

        # note: need to release the lock while waiting for the thread to finish
        hyperborea.proxy.device_lock.release()
        self.scan_thread.join()
        hyperborea.proxy.device_lock.acquire()

        self.remote.close()
        self.status_pipe.close()
        self.control_pipe.close()

    def do_active_scan(self, request):
        sn, is_boot = request
        try:
            if not is_boot:
                self.device.connect_radio(sn)
            else:
                self.device.connect_radio_boot(sn)

            self.remote.wait_for_connect(1000)

            device_info = hyperborea.device_info.get_active_scan_info(
                self.device, self.remote, self.diskcache)

            t1 = device_info['user_tag_1']
            t2 = device_info['user_tag_2']
            board_info = device_info['board_info']
            build_info = device_info['build_info']
            build_date = device_info['build_date']
            has_bootloader = bool(device_info['bootloader_info'])

            last_seen = datetime.datetime.now(datetime.timezone.utc)
            s = ScanResult(True, sn, is_boot, None, None, last_seen, t1, t2,
                           board_info, build_info, build_date, has_bootloader)
            self.status_pipe.send(s)
        except Exception:
            # couldn't do the active scan
            # no big deal
            pass
        finally:
            self.device.stop_radio()

    def collect_scan_results(self):
        last_seen = datetime.datetime.now(datetime.timezone.utc)

        # get the results
        results = self.device.get_radio_extra_scan_results()

        # get the scan powers
        scan_powers = {}  # key: sn, value: power dBm
        if self.supports_scan_power:
            try:
                for i in range(0, len(results), self.power_max_queries):
                    result_subset = results[i:i + self.power_max_queries]
                    serials = [r.serial_number for r in result_subset]
                    powers = self.device.get_radio_scan_power(serials)
                    for sn, power in zip(serials, powers):
                        if power != 0x7F:
                            scan_powers[sn] = power
            except asphodel.AsphodelError as e:
                if e.args[1] == "ERROR_CODE_UNIMPLEMENTED_COMMAND":
                    self.supports_scan_power = False
                else:
                    raise

        for r in results:
            power = scan_powers.get(r.serial_number, None)
            is_bootloader = bool(r.asphodel_type &
                                 asphodel.ASPHODEL_PROTOCOL_TYPE_BOOTLOADER)
            s = ScanResult(None, r.serial_number, is_bootloader, r.device_mode,
                           power, last_seen, None, None, None, None, None,
                           None)
            self.status_pipe.send(s)

    def scan_thread_run(self):
        try:
            active_scans = collections.deque()

            with hyperborea.proxy.device_lock:
                self.device.start_radio_scan()

            # next allowable time to stop the scan
            scan_stop = time.time() + 0.6

            restart_scan = False

            while not self.finished.is_set():
                if restart_scan:
                    with hyperborea.proxy.device_lock:
                        self.device.start_radio_scan()
                    restart_scan = False

                # poll control pipe
                while True:
                    if self.control_pipe.poll():
                        try:
                            request = self.control_pipe.recv()
                            active_scans.append(request)
                        except EOFError:
                            # pipe broken, treat it like a stop call
                            self.finished.set()
                            break
                    else:
                        break

                if self.finished.is_set():
                    with hyperborea.proxy.device_lock:
                        self.device.stop_radio()
                    break

                # collect all scan results from the device
                with hyperborea.proxy.device_lock:
                    self.collect_scan_results()

                if time.time() < scan_stop:
                    # not enough time scanning
                    continue

                # do all active scan connects
                while not self.finished.is_set():
                    try:
                        request = active_scans.popleft()
                    except IndexError:
                        # no (more) devices
                        break

                    with hyperborea.proxy.device_lock:
                        self.do_active_scan(request)
                    restart_scan = True

                if self.finished.is_set():
                    with hyperborea.proxy.device_lock:
                        self.device.stop_radio()
                    break

                # do a boot scan
                with hyperborea.proxy.device_lock:
                    self.device.start_radio_scan_boot()
                    end_time = time.time() + 0.1
                    while time.time() < end_time:
                        if self.finished.is_set():
                            break
                        self.collect_scan_results()
                restart_scan = True
                scan_stop = time.time() + 0.6
        except Exception:
            self.logger.exception("Unhandled exception in scan_thread_run")
            self.status_pipe.send("error")


class RadioScanTableModel(QtCore.QAbstractTableModel):
    def __init__(self, parent=None):
        super().__init__(parent)

        self.filter_level = 0

        self.scan_list = []

        self.header = ["Serial", "Strength", "Tag 1", "Tag 2",
                       "Board Info", "Build Info", "Build Date", "Bootloader",
                       "Device Mode", "Last Seen"]

        self.update_timer = QtCore.QTimer(self)
        self.update_timer.timeout.connect(self.update_timer_cb)

        self.last_sort_col = 0
        self.last_sort_reverse = False

    def rowCount(self, parent):
        return len(self.scan_list)

    def columnCount(self, parent):
        return len(self.header)

    def data(self, index, role):
        if not index.isValid():
            return None

        scan_result = self.scan_list[index.row()]
        col = index.column()

        if role != QtCore.Qt.DisplayRole:
            if col == 9:
                now = datetime.datetime.now(datetime.timezone.utc)
                delta = now - scan_result.last_seen
                seconds = int(delta.total_seconds())
                if seconds > 30:
                    if role == QtCore.Qt.BackgroundRole:
                        return QtGui.QColor(QtCore.Qt.red)
                    elif role == QtCore.Qt.ForegroundRole:
                        return QtGui.QColor(QtCore.Qt.black)
                elif seconds > 10:
                    if role == QtCore.Qt.BackgroundRole:
                        return QtGui.QColor(QtCore.Qt.yellow)
                    elif role == QtCore.Qt.ForegroundRole:
                        return QtGui.QColor(QtCore.Qt.black)

            if scan_result.is_bootloader:
                if role == QtCore.Qt.BackgroundRole:
                    return QtGui.QColor(QtCore.Qt.yellow)
                elif role == QtCore.Qt.ForegroundRole:
                    return QtGui.QColor(QtCore.Qt.black)
                elif role == QtCore.Qt.FontRole and col == 7:
                    font = QtGui.QFont()
                    font.setBold(True)
                    return font

            # default for non-display roles
            return None

        # must be a display role (i.e. return the string contents)

        if col == 0:  # Serial Number
            return str(scan_result.serial_number)
        elif col == 1:  # Radio Strength
            if scan_result.scan_strength is not None:
                return "{} dBm".format(scan_result.scan_strength)
        elif col == 2:  # User Tag 1
            if scan_result.user_tag_1:
                return scan_result.user_tag_1
        elif col == 3:  # User Tag 2
            if scan_result.user_tag_2:
                return scan_result.user_tag_2
        elif col == 4:  # Board Info
            if scan_result.board_info:
                return "{} rev {}".format(*scan_result.board_info)
        elif col == 5:  # Build Info
            if scan_result.build_info:
                return scan_result.build_info
        elif col == 6:  # Build Info
            if scan_result.build_date:
                return scan_result.build_date
        elif col == 7:  # Bootloader
            if scan_result.is_bootloader:
                return "Running"
            elif scan_result.has_bootloader is True:
                return "Capable"
            elif scan_result.has_bootloader is False:
                return "Not Capable"
        elif col == 8:  # Device Mode
            return str(scan_result.device_mode)
        else:  # Last Seen
            now = datetime.datetime.now(datetime.timezone.utc)
            delta = now - scan_result.last_seen
            seconds = int(delta.total_seconds())
            return str("{}s ago".format(seconds))
        return None

    def flags(self, index):
        if not index.isValid():
            return None

        scan_result = self.scan_list[index.row()]
        if scan_result.scan_strength is None:
            # radio doesn't support scan strength
            return QtCore.Qt.ItemIsSelectable | QtCore.Qt.ItemIsEnabled
        elif scan_result.scan_strength < self.filter_level:
            # too low, don't enable
            return QtCore.Qt.ItemIsSelectable
        else:
            return QtCore.Qt.ItemIsSelectable | QtCore.Qt.ItemIsEnabled

    def headerData(self, col, orientation, role):
        if orientation == QtCore.Qt.Horizontal:
            if role == QtCore.Qt.DisplayRole:
                return self.header[col]
        return None

    def get_sort_key(self, col):
        def get_key(attrs):
            def f(scan_result):
                values = (getattr(scan_result, attr) for attr in attrs)
                comparable_values = ((v is None, v) for v in values)

                if scan_result.scan_strength is None:
                    filtered = False
                else:
                    filtered = scan_result.scan_strength < self.filter_level

                return [filtered, tuple(comparable_values)]
            return f

        if col == 0:  # Serial Number
            return get_key(["serial_number"])
        elif col == 1:  # Radio Strength
            return get_key(["scan_strength"])
        elif col == 2:  # User Tag 1
            return get_key(['user_tag_1'])
        elif col == 3:  # User Tag 2
            return get_key(['user_tag_2'])
        elif col == 4:  # Board Info
            return get_key(['board_info'])
        elif col == 5:  # Build Info
            return get_key(['build_info'])
        elif col == 6:  # Build Date
            return get_key(['build_date'])
        elif col == 7:  # Bootloader
            return get_key(['is_bootloader', 'has_bootloader'])
        elif col == 8:  # Device Mode
            return get_key(['device_mode'])
        else:  # Last Seen
            return get_key(['last_seen'])

    def sort(self, col, order):
        """sort table by given column number col"""
        self.layoutAboutToBeChanged.emit()

        reverse = (order == QtCore.Qt.DescendingOrder)

        self.last_sort_col = col
        self.last_sort_reverse = reverse

        self.scan_list.sort(key=self.get_sort_key(col), reverse=reverse)

        self.layoutChanged.emit()

    def create_updated_result(self, original, new):
        d = original._asdict()
        for key, value in new._asdict().items():
            if value is not None:
                d[key] = value
        return ScanResult(**d)

    def insert_scan_result(self, scan_result):
        if not self.scan_list:
            # first entry
            self.update_timer.start(1000)

        original_index = None
        for i, s in enumerate(self.scan_list):
            if s.serial_number == scan_result.serial_number:
                # have a duplicate
                original_index = i
                scan_result = self.create_updated_result(s, scan_result)
                break

        key = self.get_sort_key(self.last_sort_col)
        keys = [key(s) for s in self.scan_list]

        if original_index is not None:
            del keys[original_index]

        new_index = bisect.bisect_right(keys, key(scan_result))

        parent = QtCore.QModelIndex()
        if original_index is not None:
            if original_index == new_index:
                # just an update
                top_left_index = self.index(original_index, 0)
                bottom_right_index = self.index(original_index,
                                                len(self.header) - 1)
                self.scan_list[original_index] = scan_result
                self.dataChanged.emit(top_left_index, bottom_right_index)
            else:
                if new_index < original_index:
                    move_index = new_index
                else:
                    move_index = new_index + 1
                self.beginMoveRows(parent, original_index, original_index,
                                   parent, move_index)
                del self.scan_list[original_index]
                self.scan_list.insert(new_index, scan_result)
                self.endMoveRows()
        else:
            self.beginInsertRows(parent, new_index, new_index)
            self.scan_list.insert(new_index, scan_result)
            self.endInsertRows()

    def mark_for_rescan(self, scan_result):
        index = None
        for i, s in enumerate(self.scan_list):
            if s.serial_number == scan_result.serial_number:
                index = i
                break

        d = s._asdict()
        d['active_scan'] = None

        top_left_index = self.index(index, 0)
        bottom_right_index = self.index(index, len(self.header) - 1)
        self.scan_list[index] = ScanResult(**d)
        self.dataChanged.emit(top_left_index, bottom_right_index)

    def has_active_scan(self, serial_number):
        for s in self.scan_list:
            if s.serial_number == serial_number:
                return (s.active_scan is True)

    def clear_scan_results(self):
        if self.scan_list:
            self.layoutAboutToBeChanged.emit()
            self.scan_list.clear()
            self.layoutChanged.emit()
        self.update_timer.stop()

    def update_timer_cb(self):
        start_index = self.index(0, 7)
        end_index = self.index(len(self.scan_list), 7)
        self.dataChanged.emit(start_index, end_index)

    def set_filter_level(self, new_filter_level):
        filtered_by_old = set()
        filtered_by_new = set()
        for scan_result in self.scan_list:
            if scan_result.scan_strength < self.filter_level:
                filtered_by_old.add(scan_result)
            if scan_result.scan_strength < new_filter_level:
                filtered_by_new.add(scan_result)

        if filtered_by_new != filtered_by_old:
            self.layoutAboutToBeChanged.emit()
            self.filter_level = new_filter_level
            self.scan_list.sort(key=self.get_sort_key(self.last_sort_col),
                                reverse=self.last_sort_reverse)
            self.layoutChanged.emit()
        else:
            self.filter_level = new_filter_level


class RadioScanDialog(QtWidgets.QDialog, Ui_RadioScanDialog):
    def __init__(self, parent=None):
        super().__init__(parent)

        self.status_pipe = None
        self.control_pipe = None

        self.settings = QtCore.QSettings()

        self.setupUi(self)
        self.extra_ui_setup()

        self.status_timer = QtCore.QTimer(self)
        self.status_timer.timeout.connect(self.status_timer_cb)

        self.selection_changed()

    def extra_ui_setup(self):
        self.table_model = RadioScanTableModel(self)
        self.tableView.setModel(self.table_model)

        header = self.tableView.horizontalHeader()
        header.setResizeMode(QtWidgets.QHeaderView.ResizeToContents)

        self.clearButton = self.buttonBox.button(
            QtWidgets.QDialogButtonBox.Reset)
        self.clearButton.setText(self.tr("Clear"))
        self.clearButton.clicked.connect(self.table_model.clear_scan_results)

        self.filterSlider.valueChanged.connect(self.set_filter_level)
        self.activeScan.toggled.connect(self.set_active_scan)

        selection_model = self.tableView.selectionModel()
        selection_model.selectionChanged.connect(self.selection_changed)
        self.tableView.doubleClicked.connect(self.double_click_cb)

    def double_click_cb(self, index=None):
        self.accept()

    def get_selected_scan_result(self):
        selected_devices = []
        for index in self.tableView.selectionModel().selectedRows():
            selected_devices.append(self.table_model.scan_list[index.row()])
        return selected_devices[0]

    def selection_changed(self, selected=None, deselected=None):
        ok_button = self.buttonBox.button(QtWidgets.QDialogButtonBox.Ok)
        if self.tableView.selectionModel().hasSelection():
            ok_button.setEnabled(True)
        else:
            ok_button.setEnabled(False)

    def status_timer_cb(self):
        while self.status_pipe.poll():
            try:
                message = self.status_pipe.recv()
                if isinstance(message, ScanResult):
                    scan_result = message
                    self.table_model.insert_scan_result(scan_result)

                    if self.activeScan.isChecked():
                        has_active_scan = self.table_model.has_active_scan(
                            scan_result.serial_number)
                        if not has_active_scan:
                            self.control_pipe.send((scan_result.serial_number,
                                                    scan_result.is_bootloader))
                else:
                    # error
                    self.stop_scan()
                    return
            except EOFError:
                break

    def set_filter_level(self, value):
        self.table_model.set_filter_level(value)
        self.settings.setValue("ScanFilterLevel", value)

    def set_active_scan(self, junk=None):
        value = 1 if self.activeScan.isChecked() else 0
        self.settings.setValue("ActiveScan", value)

    def start_scan(self, status_pipe, control_pipe):
        # read check boxes and filter slider states from settings
        active_scan = read_bool_setting(self.settings, "ActiveScan", False)
        self.activeScan.setChecked(active_scan)
        filter_level = self.settings.value("ScanFilterLevel")
        if filter_level is not None:
            try:
                filter_level = int(filter_level)
            except ValueError:
                filter_level = self.filterSlider.minimum()  # default
        else:
            filter_level = self.filterSlider.minimum()  # default
        self.filterSlider.setValue(filter_level)

        self.status_pipe = status_pipe
        self.control_pipe = control_pipe
        self.status_timer.start(20)  # 20 milliseconds

    def stop_scan(self):
        self.status_timer.stop()
        self.status_pipe = None
        self.control_pipe = None

    def mark_for_rescan(self, scan_result):
        self.table_model.mark_for_rescan(scan_result)



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'remote_panel.ui'
##
## Created by: Qt User Interface Compiler version 5.15.2
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide2.QtCore import *
from PySide2.QtGui import *
from PySide2.QtWidgets import *


class Ui_RemotePanel(object):
    def setupUi(self, RemotePanel):
        if not RemotePanel.objectName():
            RemotePanel.setObjectName(u"RemotePanel")
        RemotePanel.resize(112, 182)
        self.verticalLayout = QVBoxLayout(RemotePanel)
        self.verticalLayout.setObjectName(u"verticalLayout")
        self.verticalSpacer = QSpacerItem(10, 10, QSizePolicy.Minimum, QSizePolicy.Expanding)

        self.verticalLayout.addItem(self.verticalSpacer)

        self.goToParentButton = QPushButton(RemotePanel)
        self.goToParentButton.setObjectName(u"goToParentButton")

        self.verticalLayout.addWidget(self.goToParentButton)


        self.retranslateUi(RemotePanel)

        QMetaObject.connectSlotsByName(RemotePanel)
    # setupUi

    def retranslateUi(self, RemotePanel):
        RemotePanel.setWindowTitle(QCoreApplication.translate("RemotePanel", u"Remote Device Panel", None))
        RemotePanel.setTitle(QCoreApplication.translate("RemotePanel", u"Remote Device", None))
        self.goToParentButton.setText(QCoreApplication.translate("RemotePanel", u"Go To Radio Tab", None))
    # retranslateUi




# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'ctrl_var_widget.ui'
##
## Created by: Qt User Interface Compiler version 5.15.2
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide2.QtCore import *
from PySide2.QtGui import *
from PySide2.QtWidgets import *


class Ui_CtrlVarWidget(object):
    def setupUi(self, CtrlVarWidget):
        if not CtrlVarWidget.objectName():
            CtrlVarWidget.setObjectName(u"CtrlVarWidget")
        CtrlVarWidget.resize(121, 19)
        self.horizontalLayout = QHBoxLayout(CtrlVarWidget)
        self.horizontalLayout.setContentsMargins(0, 0, 0, 0)
        self.horizontalLayout.setObjectName(u"horizontalLayout")
        self.nameLabel = QLabel(CtrlVarWidget)
        self.nameLabel.setObjectName(u"nameLabel")

        self.horizontalLayout.addWidget(self.nameLabel)

        self.slider = QSlider(CtrlVarWidget)
        self.slider.setObjectName(u"slider")
        self.slider.setMinimumSize(QSize(50, 0))
        self.slider.setOrientation(Qt.Horizontal)

        self.horizontalLayout.addWidget(self.slider)


        self.retranslateUi(CtrlVarWidget)

        QMetaObject.connectSlotsByName(CtrlVarWidget)
    # setupUi

    def retranslateUi(self, CtrlVarWidget):
        CtrlVarWidget.setWindowTitle(QCoreApplication.translate("CtrlVarWidget", u"Ctrl Var Widget", None))
        self.nameLabel.setText(QCoreApplication.translate("CtrlVarWidget", u"Ctrl Var Name", None))
    # retranslateUi




# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'radio_scan_dialog.ui'
##
## Created by: Qt User Interface Compiler version 5.15.2
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide2.QtCore import *
from PySide2.QtGui import *
from PySide2.QtWidgets import *


class Ui_RadioScanDialog(object):
    def setupUi(self, RadioScanDialog):
        if not RadioScanDialog.objectName():
            RadioScanDialog.setObjectName(u"RadioScanDialog")
        RadioScanDialog.resize(1000, 445)
        self.verticalLayout = QVBoxLayout(RadioScanDialog)
        self.verticalLayout.setObjectName(u"verticalLayout")
        self.tableView = QTableView(RadioScanDialog)
        self.tableView.setObjectName(u"tableView")
        self.tableView.setVerticalScrollBarPolicy(Qt.ScrollBarAlwaysOn)
        self.tableView.setEditTriggers(QAbstractItemView.NoEditTriggers)
        self.tableView.setSelectionMode(QAbstractItemView.SingleSelection)
        self.tableView.setSelectionBehavior(QAbstractItemView.SelectRows)
        self.tableView.setHorizontalScrollMode(QAbstractItemView.ScrollPerPixel)
        self.tableView.setSortingEnabled(True)
        self.tableView.setCornerButtonEnabled(False)
        self.tableView.horizontalHeader().setStretchLastSection(True)
        self.tableView.verticalHeader().setVisible(False)

        self.verticalLayout.addWidget(self.tableView)

        self.horizontalLayout = QHBoxLayout()
        self.horizontalLayout.setObjectName(u"horizontalLayout")
        self.activeScan = QCheckBox(RadioScanDialog)
        self.activeScan.setObjectName(u"activeScan")

        self.horizontalLayout.addWidget(self.activeScan)

        self.horizontalSpacer = QSpacerItem(40, 20, QSizePolicy.MinimumExpanding, QSizePolicy.Minimum)

        self.horizontalLayout.addItem(self.horizontalSpacer)

        self.filterLabel = QLabel(RadioScanDialog)
        self.filterLabel.setObjectName(u"filterLabel")

        self.horizontalLayout.addWidget(self.filterLabel)

        self.filterSlider = QSlider(RadioScanDialog)
        self.filterSlider.setObjectName(u"filterSlider")
        sizePolicy = QSizePolicy(QSizePolicy.Fixed, QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.filterSlider.sizePolicy().hasHeightForWidth())
        self.filterSlider.setSizePolicy(sizePolicy)
        self.filterSlider.setMinimumSize(QSize(200, 0))
        self.filterSlider.setMinimum(-100)
        self.filterSlider.setMaximum(0)
        self.filterSlider.setOrientation(Qt.Horizontal)

        self.horizontalLayout.addWidget(self.filterSlider)

        self.filterSpinBox = QSpinBox(RadioScanDialog)
        self.filterSpinBox.setObjectName(u"filterSpinBox")
        self.filterSpinBox.setMinimum(-100)
        self.filterSpinBox.setMaximum(0)

        self.horizontalLayout.addWidget(self.filterSpinBox)


        self.verticalLayout.addLayout(self.horizontalLayout)

        self.buttonBox = QDialogButtonBox(RadioScanDialog)
        self.buttonBox.setObjectName(u"buttonBox")
        self.buttonBox.setOrientation(Qt.Horizontal)
        self.buttonBox.setStandardButtons(QDialogButtonBox.Cancel|QDialogButtonBox.Ok|QDialogButtonBox.Reset)

        self.verticalLayout.addWidget(self.buttonBox)


        self.retranslateUi(RadioScanDialog)
        self.buttonBox.accepted.connect(RadioScanDialog.accept)
        self.buttonBox.rejected.connect(RadioScanDialog.reject)
        self.filterSlider.valueChanged.connect(self.filterSpinBox.setValue)
        self.filterSpinBox.valueChanged.connect(self.filterSlider.setValue)

        QMetaObject.connectSlotsByName(RadioScanDialog)
    # setupUi

    def retranslateUi(self, RadioScanDialog):
        RadioScanDialog.setWindowTitle(QCoreApplication.translate("RadioScanDialog", u"Radio Scan", None))
        self.activeScan.setText(QCoreApplication.translate("RadioScanDialog", u"Active Scan", None))
        self.filterLabel.setText(QCoreApplication.translate("RadioScanDialog", u"Filter", None))
        self.filterSpinBox.setSuffix(QCoreApplication.translate("RadioScanDialog", u" dBm", None))
    # retranslateUi




import enum
import functools
import logging
import multiprocessing
import time

from PySide2 import QtCore, QtWidgets

import hyperborea.proxy

from . import bootloader
from . import radio_scan
from .bulk_update_firmware import BulkUpdateFirmwareDialog
from .ui_radio_panel import Ui_RadioPanel

logger = logging.getLogger(__name__)


def finish_scan(device):
    device.stop_radio()
    results = device.get_radio_scan_results()
    return results


def create_remote(device):
    remote = device.get_remote_device()
    remote.open()
    return remote


def bulk_claim(device):
    device_logger = hyperborea.proxy.get_device_logger(logger, device)
    remote = device.get_remote_device()
    remote.open()

    try:
        to_claim = set()

        device.start_radio_scan()
        finish = time.monotonic() + 1
        while finish > time.monotonic():
            for result in device.get_radio_extra_scan_results():
                if result.device_mode:
                    to_claim.add(result.serial_number)
        device.stop_radio()

        if not to_claim:
            device_logger.info("No Devices to Claim")
        else:
            for serial_number in to_claim:
                try:
                    device_logger.info("Claiming %s...", serial_number)
                    device.connect_radio(serial_number)
                    remote.wait_for_connect(5000)
                    remote.set_device_mode(0)
                    device_logger.info("Claimed %s", serial_number)
                except Exception:
                    device_logger.error("Error Claiming %s", serial_number)
            device_logger.info("Finished Bulk Claim")
    finally:
        device.stop_radio()
        remote.close()


class PipeWithSerial:
    def __init__(self, pipe):
        self.pipe = pipe
        self.serial_number = 0

    def send(self, message):
        if isinstance(message, str):
            new_message = "{}: {}".format(self.serial_number, message)
            self.pipe.send(new_message)
        else:
            self.pipe.send(message)

    def close(self):
        # ignore close calls, the pipe will be reused
        pass


def bulk_update(device, progress_pipe, cancel_pipe, firm_data, valid_range,
                device_mode, radio_strength):
    canceled = False
    pipe_with_serial = PipeWithSerial(progress_pipe)
    device_logger = hyperborea.proxy.get_device_logger(logger, device)
    remote = device.get_remote_device()
    remote.open()

    power_max_queries = min(device.get_max_outgoing_param_length() // 4,
                            device.get_max_incoming_param_length())

    scanned = 0
    to_program = 0
    ignored = 0
    finished = 0
    errors = 0

    try:
        # all of these contain elements of (sn, app_mode)
        unknown_serials = set()
        reprogram_serials = set()

        scan_powers = {}  # key: sn, value: power dBm

        progress_pipe.send("Scanning for bootloaders...")
        device.start_radio_scan_boot()
        finish = time.monotonic() + 1  # 1 second
        while finish > time.monotonic():
            unknown_powers = []
            for result in device.get_radio_extra_scan_results():
                if result.serial_number in valid_range:
                    unknown_serials.add((False, result.serial_number))
                    if result.serial_number not in scan_powers:
                        unknown_powers.append(result.serial_number)
            # get the powers
            if radio_strength is not None:
                for i in range(0, len(unknown_powers), power_max_queries):
                    serials = unknown_powers[i:i + power_max_queries]
                    powers = device.get_radio_scan_power(serials)
                    for sn, power in zip(serials, powers):
                        if power != 0x7F:
                            scan_powers[sn] = power
        device.stop_radio()

        progress_pipe.send("Scanning for devices...")
        device.start_radio_scan()
        finish = time.monotonic() + 2  # 2 seconds
        while finish > time.monotonic():
            unknown_powers = []
            for result in device.get_radio_extra_scan_results():
                if result.serial_number in valid_range:
                    unknown_serials.add((True, result.serial_number))
                    if result.serial_number not in scan_powers:
                        unknown_powers.append(result.serial_number)
            # get the powers
            if radio_strength is not None:
                for i in range(0, len(unknown_powers), power_max_queries):
                    serials = unknown_powers[i:i + power_max_queries]
                    powers = device.get_radio_scan_power(serials)
                    for sn, power in zip(serials, powers):
                        if power != 0x7F:
                            scan_powers[sn] = power
        device.stop_radio()

        # filter by radio strength
        if radio_strength is not None:
            for app_mode, serial_number in unknown_serials.copy():
                if serial_number in scan_powers:
                    if scan_powers[serial_number] < radio_strength:
                        unknown_serials.remove((app_mode, serial_number))

        if not unknown_serials:
            device_logger.info("No Devices Found")
            return

        scanned = len(unknown_serials)

        for app_mode, serial_number in sorted(unknown_serials):
            try:
                # connect
                progress_pipe.send("Querying {}...".format(serial_number))
                if app_mode:
                    device.connect_radio(serial_number)
                else:
                    device.connect_radio_boot(serial_number)
                remote.wait_for_connect(5000)

                if app_mode and remote.get_bootloader_info() != "Asphodel":
                    device_logger.info(
                        "{} doesn't have a bootloader".format(serial_number))
                    continue

                # see if the firmware is supported
                board_info = remote.get_board_info()
                for rev, board_name in firm_data['board']:
                    if (rev == board_info[1] and
                            board_name == board_info[0]):
                        break
                else:
                    device_logger.info("{} isn't a supported board".format(
                        serial_number))
                    continue

                if firm_data['chip'] != remote.get_chip_model():
                    device_logger.info("{} doesn't have correct chip".format(
                        serial_number))
                    continue

                in_bootloader = remote.supports_bootloader_commands()
                device_info = {'supports_bootloader': in_bootloader,
                               'build_info': remote.get_build_info(),
                               'build_date': remote.get_build_date()}
                if bootloader.already_programmed(firm_data, device_info):
                    device_logger.info("{} already programmed".format(
                        serial_number))
                    continue

                reprogram_serials.add((app_mode, serial_number))
            except Exception:
                device_logger.info("Error Querying {}".format(serial_number))

        if not reprogram_serials:
            device_logger.info("No Valid Devices Found")
            return

        to_program = len(reprogram_serials)
        ignored = scanned - to_program

        device_logger.info(
            "Bulk Update: scanned=%s, ignored=%s, to_program=%s", scanned,
            ignored, to_program)

        for app_mode, serial_number in sorted(reprogram_serials):
            if cancel_pipe.poll():
                progress_pipe.send("Cancelling...")
                canceled = True
                break

            try:
                tries = 2
                while True:
                    try:
                        # connect
                        progress_pipe.send((0, 0))
                        progress_pipe.send("Connecting to {}...".format(
                            serial_number))
                        if app_mode:
                            device.connect_radio(serial_number)
                        else:
                            device.connect_radio_boot(serial_number)
                        remote.wait_for_connect(5000)

                        # bootloader
                        pipe_with_serial.serial_number = serial_number
                        bootloader.do_bootload(remote, firm_data,
                                               pipe_with_serial)

                        # device mode
                        if device_mode is not None:
                            remote.set_device_mode(device_mode)
                    except Exception:
                        tries -= 1
                        if tries == 0:
                            raise
                        else:
                            continue
                    break
                finished += 1
                device_logger.info("Finished {}".format(serial_number))
            except Exception:
                errors += 1
                device_logger.exception("Error Programming {}".format(
                    serial_number))
    finally:
        progress_pipe.close()
        device.stop_radio()
        remote.close()

    device_logger.info(
        "Bulk Update: scanned=%s, ignored=%s, finished=%s, errors=%s", scanned,
        ignored, finished, errors)

    if canceled:
        device_logger.info("Canceled Bulk Update")
    else:
        device_logger.info("Finished Bulk Update")


@enum.unique
class RadioState(enum.Enum):
    UNINITIALIZED = 0
    DISCONNECTED = 1
    CONNECTING = 2
    CONNECTED = 3
    SCANNING = 4


class RadioPanel(QtWidgets.QGroupBox, Ui_RadioPanel):
    def __init__(self, start_usb_operation, diskcache, parent=None):
        super().__init__(parent)

        self.diskcache = diskcache
        self.device_tab = parent
        self.logger = parent.logger
        self.start_usb_operation = start_usb_operation

        self.radio_scan_dialog = None

        self.state = RadioState.UNINITIALIZED

        self.last_connect_op = None
        self.last_connect_sn = None

        self.no_streaming = False

        self.connect_available = False
        self.remote_tab = None

        self.setupUi(self)
        self.extra_ui_setup()

        self.setup_callbacks()
        self.setup_usb_operations()

    def extra_ui_setup(self):
        self.menu = QtWidgets.QMenu(self)
        self.menu.addAction(self.actionConnectNoStreaming)
        self.menu.addSeparator()
        self.menu.addAction(self.actionBulkClaim)
        self.menu.addAction(self.actionBulkUpdateFirmware)
        self.menu.addSeparator()
        self.menu.addAction(self.actionConnectAnyBootloader)
        self.menu.addAction(self.actionConnectSpecificBootloader)
        self.menu.addAction(self.actionBootloaderScan)
        self.advancedMenuButton.setMenu(self.menu)

        self.bulk_update_progress = QtWidgets.QProgressDialog(
            self.tr(""), self.tr("Cancel"), 0, 100)
        self.bulk_update_progress.setWindowTitle(self.tr("Bulk Update"))
        self.bulk_update_progress.setWindowModality(QtCore.Qt.WindowModal)
        self.bulk_update_progress.setMinimumDuration(0)
        self.bulk_update_progress.setAutoReset(False)
        self.bulk_update_progress.canceled.connect(self.bulk_update_canceled)
        self.bulk_update_progress.reset()

    def setup_callbacks(self):
        self.devicesComboBox.currentIndexChanged.connect(self.combo_changed)
        self.detailScanButton.clicked.connect(self.detail_scan)
        self.scanButton.clicked.connect(self.start_scan)
        self.connectButton.clicked.connect(self.connect_button_cb)
        self.disconnectButton.clicked.connect(self.disconnect_button_cb)
        self.goToRemoteButton.clicked.connect(self.show_remote_tab)

        self.scan_timer = QtCore.QTimer(self)
        self.scan_timer.timeout.connect(self.scan_timer_cb)
        self.connect_check_timer = QtCore.QTimer(self)
        self.connect_check_timer.timeout.connect(self.connect_check_timer_cb)

        self.actionConnectNoStreaming.triggered.connect(
            self.connect_no_streaming_cb)
        self.actionConnectAnyBootloader.triggered.connect(
            self.connect_any_bootloader_cb)
        self.actionConnectSpecificBootloader.triggered.connect(
            self.connect_specific_bootloader_cb)
        self.actionBootloaderScan.triggered.connect(self.start_scan_boot)
        self.actionBulkClaim.triggered.connect(self.start_bulk_claim)
        self.actionBulkUpdateFirmware.triggered.connect(
            self.start_bulk_update_firmware)

        self.bulk_update_progress_timer = QtCore.QTimer(self)
        self.bulk_update_progress_timer.timeout.connect(
            self.bulk_update_progress_cb)

    def setup_usb_operations(self):
        self.start_scan_op = hyperborea.proxy.SimpleDeviceOperation(
            "start_radio_scan")
        self.start_scan_op.completed.connect(self.start_scan_cb)
        self.start_scan_boot_op = hyperborea.proxy.SimpleDeviceOperation(
            "start_radio_scan_boot")
        self.start_scan_boot_op.completed.connect(self.start_boot_scan_cb)
        self.stop_op = hyperborea.proxy.SimpleDeviceOperation("stop_radio")
        self.finish_scan_op = hyperborea.proxy.DeviceOperation(finish_scan)
        self.finish_scan_op.completed.connect(self.finish_scan_cb)

        self.connect_radio_op = hyperborea.proxy.SimpleDeviceOperation(
            "connect_radio")
        self.connect_radio_op.completed.connect(self.connect_radio_cb)
        self.connect_radio_boot_op = hyperborea.proxy.SimpleDeviceOperation(
            "connect_radio_boot")
        self.connect_radio_boot_op.completed.connect(self.connect_radio_cb)
        self.get_status_op = hyperborea.proxy.SimpleDeviceOperation(
            "get_radio_status")
        self.get_status_op.completed.connect(self.get_status_cb)

        self.start_detail_scan_op = hyperborea.proxy.DeviceOperation(
            radio_scan.start_radio_scan_manager)
        self.stop_detail_scan_op = hyperborea.proxy.DeviceOperation(
            radio_scan.stop_radio_scan_manager)
        self.stop_detail_scan_op.completed.connect(self.stop_detail_scan_cb)

        self.bulk_claim_op = hyperborea.proxy.DeviceOperation(bulk_claim)
        self.bulk_claim_op.completed.connect(self.bulk_claim_cb)
        self.bulk_update_op = hyperborea.proxy.DeviceOperation(bulk_update)
        self.bulk_update_op.completed.connect(self.bulk_update_cb)

    def add_ctrl_var_widget(self, widget):
        self.ctrlVarLayout.addWidget(widget)

    def do_start_operation(self, operation, sn):
        self.last_connect_op = operation
        self.last_connect_sn = sn
        self.start_usb_operation(operation, sn)

    def do_stop_operation(self, operation):
        self.last_connect_op = None
        self.last_connect_sn = None
        self.start_usb_operation(operation)

    def go_to_disconnected_state(self):
        # no device is connected; not scanning; may have scan results
        self.state = RadioState.DISCONNECTED

        self.connect_available = True

        self.devicesComboBox.setEnabled(True)
        self.detailScanButton.setEnabled(True)
        self.scanButton.setEnabled(True)
        self.actionBootloaderScan.setEnabled(True)
        self.disconnectButton.setEnabled(False)
        self.goToRemoteButton.setEnabled(False)
        self.actionConnectAnyBootloader.setEnabled(True)
        self.actionConnectSpecificBootloader.setEnabled(True)
        self.actionBulkClaim.setEnabled(True)
        self.actionBulkUpdateFirmware.setEnabled(True)

        self.scanButton.setText(self.tr("Scan"))
        self.connectButton.setText(self.tr("Connect"))

        if self.devicesComboBox.currentText() == "":
            self.connectButton.setEnabled(False)
            self.actionConnectNoStreaming.setEnabled(False)
        else:
            self.connectButton.setEnabled(True)
            self.actionConnectNoStreaming.setEnabled(True)

        self.device_tab.rgb_remote_disconnected()

    def go_to_connecting_state(self):
        # trying to connect to a device
        self.state = RadioState.CONNECTING

        self.connect_available = False

        self.devicesComboBox.setEnabled(True)
        self.detailScanButton.setEnabled(False)
        self.scanButton.setEnabled(False)
        self.actionBootloaderScan.setEnabled(False)
        self.connectButton.setEnabled(False)
        self.disconnectButton.setEnabled(True)
        self.goToRemoteButton.setEnabled(False)
        self.actionConnectNoStreaming.setEnabled(False)
        self.actionConnectAnyBootloader.setEnabled(False)
        self.actionConnectSpecificBootloader.setEnabled(False)
        self.actionBulkClaim.setEnabled(False)
        self.actionBulkUpdateFirmware.setEnabled(False)

        self.scanButton.setText(self.tr("Scan"))
        self.connectButton.setText(self.tr("Connecting..."))

        self.device_tab.rgb_remote_disconnected()

    def go_to_connected_state(self):
        # device is connected
        self.state = RadioState.CONNECTED

        self.connect_available = False

        self.devicesComboBox.setEnabled(True)
        self.detailScanButton.setEnabled(False)
        self.scanButton.setEnabled(False)
        self.actionBootloaderScan.setEnabled(False)
        self.connectButton.setEnabled(False)
        self.disconnectButton.setEnabled(True)
        self.goToRemoteButton.setEnabled(True)
        self.actionConnectNoStreaming.setEnabled(False)
        self.actionConnectAnyBootloader.setEnabled(False)
        self.actionConnectSpecificBootloader.setEnabled(False)
        self.actionBulkClaim.setEnabled(False)
        self.actionBulkUpdateFirmware.setEnabled(False)

        self.scanButton.setText(self.tr("Scan"))
        self.connectButton.setText(self.tr("Connected"))

        self.device_tab.rgb_remote_connected()

    def go_to_scanning_state(self):
        # scanning for devices; not connected or disconnected
        self.state = RadioState.SCANNING

        self.connect_available = False

        self.devicesComboBox.clear()
        self.devicesComboBox.setEnabled(True)
        self.detailScanButton.setEnabled(False)
        self.scanButton.setEnabled(False)
        self.actionBootloaderScan.setEnabled(False)
        self.connectButton.setEnabled(False)
        self.disconnectButton.setEnabled(False)
        self.goToRemoteButton.setEnabled(False)
        self.actionConnectNoStreaming.setEnabled(False)
        self.actionConnectAnyBootloader.setEnabled(False)
        self.actionConnectSpecificBootloader.setEnabled(False)
        self.actionBulkClaim.setEnabled(False)
        self.actionBulkUpdateFirmware.setEnabled(False)

        self.scanButton.setText(self.tr("Scanning..."))
        self.connectButton.setText(self.tr("Connect"))

        self.device_tab.rgb_remote_disconnected()

    def connected(self, device_info):
        # the device has been reconnected
        if self.state != RadioState.UNINITIALIZED:
            return  # wasn't disconnected, ignore

        self.devicesComboBox.clear()
        self.go_to_disconnected_state()
        self.has_scan_power = device_info['radio_scan_power']
        default_serial = device_info['radio_default_serial'] & 0xFFFFFFFF
        if default_serial == 0:
            self.start_scan()
        else:
            self.no_streaming = False
            self.devicesComboBox.addItem(str(default_serial))
            self.devicesComboBox.setCurrentIndex(0)
            self.do_start_operation(self.connect_radio_op, default_serial)
            self.go_to_connecting_state()

    def disconnected(self):
        # the device has been disconnected (e.g. unplugged)
        self.state = RadioState.UNINITIALIZED

        if self.radio_scan_dialog:
            self.radio_scan_dialog.reject()

        self.scan_timer.stop()
        self.connect_check_timer.stop()

        self.close_remote_tab()

        self.connect_available = False

        # disable all
        self.devicesComboBox.clear()
        self.devicesComboBox.setEnabled(False)
        self.detailScanButton.setEnabled(False)
        self.scanButton.setEnabled(False)
        self.actionBootloaderScan.setEnabled(False)
        self.connectButton.setEnabled(False)
        self.disconnectButton.setEnabled(False)
        self.goToRemoteButton.setEnabled(False)
        self.actionConnectNoStreaming.setEnabled(False)
        self.actionConnectAnyBootloader.setEnabled(False)
        self.actionConnectSpecificBootloader.setEnabled(False)
        self.actionBulkClaim.setEnabled(False)
        self.actionBulkUpdateFirmware.setEnabled(False)

        self.scanButton.setText(self.tr("Scan"))
        self.connectButton.setText(self.tr("Connect"))

    def stop(self):
        if self.radio_scan_dialog:
            self.radio_scan_dialog.reject()
        self.close_remote_tab()
        self.do_stop_operation(self.stop_op)

    def start_scan(self):
        self.go_to_scanning_state()
        self.do_stop_operation(self.start_scan_op)

    def start_scan_boot(self):
        self.go_to_scanning_state()
        self.do_stop_operation(self.start_scan_boot_op)

    def start_scan_cb(self):
        self.scan_timer.start(600)  # 0.6 seconds

    def start_boot_scan_cb(self):
        self.scan_timer.start(2000)  # 2.0 seconds

    def scan_timer_cb(self):
        self.start_usb_operation(self.finish_scan_op)
        self.scan_timer.stop()

    def finish_scan_cb(self, results):
        self.devicesComboBox.clear()
        for result in sorted(results):
            result_str = str(result)
            self.devicesComboBox.addItem(result_str)

        if results:
            self.devicesComboBox.setCurrentIndex(0)

        if self.state == RadioState.SCANNING:
            self.go_to_disconnected_state()

    def combo_changed(self, junk=None):
        if self.connect_available:
            if self.devicesComboBox.currentText() == "":
                self.connectButton.setEnabled(False)
                self.actionConnectNoStreaming.setEnabled(False)
            else:
                self.connectButton.setEnabled(True)
                self.actionConnectNoStreaming.setEnabled(True)

    def connect_button_cb(self):
        sn_str = self.devicesComboBox.currentText()
        try:
            sn = int(sn_str) & 0xFFFFFFFF
        except ValueError:
            # TODO: error box
            return
        self.no_streaming = False
        self.do_start_operation(self.connect_radio_op, sn)
        self.go_to_connecting_state()

    def connect_radio_cb(self):
        # immediately check the connected status
        self.connect_check_timer_cb()

    def connect_check_timer_cb(self):
        self.connect_check_timer.stop()  # will be restarted if necessary
        if self.state in (RadioState.CONNECTED, RadioState.CONNECTING):
            self.start_usb_operation(self.get_status_op)

    def get_status_cb(self, status):
        connected = status[0]
        if connected:
            self.device_tab.rgb_remote_connected()
            old_sn_str = self.devicesComboBox.currentText()
            new_sn_str = str(status[1])
            if old_sn_str != new_sn_str:
                # update serial number
                index = self.devicesComboBox.currentIndex()
                if index >= 0:
                    self.devicesComboBox.setItemText(index, new_sn_str)
                else:
                    self.devicesComboBox.clear()
                    self.devicesComboBox.addItem(new_sn_str)
                    self.devicesComboBox.setCurrentIndex(0)

            subproxy = self.device_tab.proxy.create_subproxy(create_remote)
            cb = functools.partial(self.subproxy_connected_cb, subproxy)
            subproxy.connected.connect(cb)
            subproxy.disconnected.connect(self.reconnect)
            subproxy.open_connection()
        elif status[1] == 0:
            # disconnected immediately; go straight to reconnect
            if self.state in (RadioState.CONNECTED, RadioState.CONNECTING):
                self.reconnect()
        else:
            self.connect_check_timer.start(250)  # 0.25 seconds

    def subproxy_connected_cb(self, subproxy):
        if self.state in (RadioState.CONNECTED, RadioState.CONNECTING):
            get_sn_op = hyperborea.proxy.SimpleDeviceOperation(
                "get_serial_number")
            cb = functools.partial(self.get_sn_cb, subproxy)
            get_sn_op.completed.connect(cb)
            subproxy.send_job(get_sn_op)
        else:
            subproxy.disconnected.disconnect(self.reconnect)
            subproxy.close_connection()

    def get_sn_cb(self, subproxy, serial):
        subproxy.disconnected.disconnect(self.reconnect)

        if self.state not in (RadioState.CONNECTED, RadioState.CONNECTING):
            subproxy.close_connection()
            return

        self.logger.info("Connected to {}".format(serial))

        if self.remote_tab:
            self.remote_tab.set_proxy(subproxy)
            self.go_to_connected_state()
        else:
            self.remote_tab = self.device_tab.plotmain.create_tab(
                subproxy, serial, [] if self.no_streaming else True)

            # make a reference to this tab so the remote panel can function
            self.remote_tab.radio_tab = self.device_tab

            # have the remote's close button act like a disconnect press
            self.remote_tab.close_pressed.connect(self.disconnect_button_cb)

            # reconnect when it disconnects
            self.remote_tab.disconnected_signal.connect(self.reconnect)

            self.go_to_connected_state()
            self.show_remote_tab()

    def disconnect_button_cb(self):
        self.scan_timer.stop()
        self.connect_check_timer.stop()
        self.close_remote_tab()
        self.do_stop_operation(self.stop_op)
        self.go_to_disconnected_state()

    def close_remote_tab(self):
        if self.remote_tab:
            self.logger.info("Closed connection to {}".format(
                self.remote_tab.serial_number))
            self.device_tab.plotmain.close_tab(self.remote_tab)
            self.remote_tab = None

    def show_remote_tab(self):
        if self.remote_tab:
            self.device_tab.plotmain.show_tab(self.remote_tab)

    def connect_any_bootloader_cb(self):
        self.no_streaming = False
        self.devicesComboBox.clear()
        self.devicesComboBox.addItem("Bootloader")
        self.devicesComboBox.setCurrentIndex(0)
        self.do_start_operation(self.connect_radio_boot_op, -1 & 0xFFFFFFFF)
        self.go_to_connecting_state()

    def connect_specific_bootloader_cb(self):
        sn, ok = QtWidgets.QInputDialog.getInt(
            None, self.tr("Bootloader Serial"),
            self.tr("Input bootloader serial number"))
        if not ok:
            return
        sn = sn & 0xFFFFFFFF

        self.no_streaming = False
        self.devicesComboBox.clear()
        self.devicesComboBox.addItem(str(sn))
        self.devicesComboBox.setCurrentIndex(0)
        self.do_start_operation(self.connect_radio_boot_op, sn)
        self.go_to_connecting_state()

    def reconnect(self):
        if self.last_connect_op:
            if self.state in (RadioState.CONNECTED, RadioState.CONNECTING):
                self.do_start_operation(self.last_connect_op,
                                        self.last_connect_sn)
                self.go_to_connecting_state()
                if self.remote_tab:
                    self.logger.info("Reconnecting to {}...".format(
                        self.remote_tab.serial_number))
                    self.remote_tab.set_disconnected("Reconnecting...")
                    self.goToRemoteButton.setEnabled(True)

    def connect_no_streaming_cb(self):
        sn_str = self.devicesComboBox.currentText()
        try:
            sn = int(sn_str) & 0xFFFFFFFF
        except ValueError:
            # TODO: error box
            return
        self.no_streaming = True
        self.do_start_operation(self.connect_radio_op, sn)
        self.go_to_connecting_state()

    def detail_scan(self):
        if self.radio_scan_dialog is None:
            self.radio_scan_dialog = radio_scan.RadioScanDialog(self)

        rx, tx = multiprocessing.Pipe(False)
        self.rx_status_pipe = rx
        self.tx_status_pipe = tx
        rx, tx = multiprocessing.Pipe(False)
        self.rx_control_pipe = rx
        self.tx_control_pipe = tx

        # start the background thread
        self.start_usb_operation(self.start_detail_scan_op,
                                 self.diskcache,
                                 self.tx_status_pipe,
                                 self.rx_control_pipe)

        # start the foreground thread
        self.radio_scan_dialog.start_scan(self.rx_status_pipe,
                                          self.tx_control_pipe)

        # run the dialog
        ret = self.radio_scan_dialog.exec_()

        # stop the background thread
        self.start_usb_operation(self.stop_detail_scan_op)

        # stop the foreground thread
        self.radio_scan_dialog.stop_scan()

        if ret == 0:
            return  # user canceled

        result = self.radio_scan_dialog.get_selected_scan_result()
        self.radio_scan_dialog.mark_for_rescan(result)

        self.devicesComboBox.clear()
        self.no_streaming = False
        self.devicesComboBox.addItem(str(result.serial_number))
        self.devicesComboBox.setCurrentIndex(0)
        if result.is_bootloader:
            self.do_start_operation(self.connect_radio_boot_op,
                                    result.serial_number)
        else:
            self.do_start_operation(self.connect_radio_op,
                                    result.serial_number)
        self.go_to_connecting_state()

    def stop_detail_scan_cb(self):
        self.rx_status_pipe.close()
        self.tx_control_pipe.close()
        self.rx_status_pipe = None
        self.tx_status_pipe = None
        self.rx_control_pipe = None
        self.tx_control_pipe = None

    def start_bulk_claim(self):
        self.go_to_scanning_state()
        self.start_usb_operation(self.bulk_claim_op)

    def bulk_claim_cb(self):
        if self.state == RadioState.SCANNING:
            self.go_to_disconnected_state()

    def start_bulk_update_firmware(self):
        dialog = BulkUpdateFirmwareDialog(self.has_scan_power,
                                          parent=self)
        ret = dialog.exec_()
        if ret == 0:
            return  # user canceled

        results = dialog.get_results()
        firm_file = results['firmware_location']

        try:
            firm_data = bootloader.decode_firm_file(firm_file)
        except Exception:
            self.logger.exception('Error loading firmware from "%s"',
                                  firm_file)
            m = self.tr('Error loading firmware from file!').format(firm_file)
            QtWidgets.QMessageBox.critical(self, self.tr("Error"), m)
            return

        rx, tx = multiprocessing.Pipe(False)
        self.bulk_update_rx_pipe = rx
        self.bulk_update_tx_pipe = tx
        rx, tx = multiprocessing.Pipe(False)
        self.cancel_rx_pipe = rx
        self.cancel_tx_pipe = tx

        # setup the progress dialog
        self.bulk_update_progress.setMinimum(0)
        self.bulk_update_progress.setMaximum(0)
        self.bulk_update_progress.setValue(0)
        self.bulk_update_progress.setLabelText(self.tr("Initializing..."))
        self.bulk_update_progress.forceShow()

        self.bulk_update_progress_timer.start(20)  # 20 milliseconds

        valid_range = range(results['min_serial'], results['max_serial'] + 1)

        self.go_to_scanning_state()
        self.start_usb_operation(
            self.bulk_update_op, self.bulk_update_tx_pipe, self.cancel_rx_pipe,
            firm_data, valid_range, results['device_mode'],
            results['radio_strength'])

    def bulk_update_canceled(self):
        if self.cancel_tx_pipe is not None:
            self.cancel_tx_pipe.send(True)

    def bulk_update_progress_cb(self):
        last_value = None
        while self.bulk_update_rx_pipe.poll():
            try:
                data = self.bulk_update_rx_pipe.recv()
                if isinstance(data, str):
                    self.bulk_update_progress.setLabelText(data)
                if isinstance(data, tuple):
                    self.bulk_update_progress.setMinimum(data[0])
                    self.bulk_update_progress.setMaximum(data[1])
                if isinstance(data, int):
                    last_value = data
            except EOFError:
                break
        if last_value is not None:
            self.bulk_update_progress.setValue(last_value)

    def bulk_update_cb(self):
        self.bulk_update_progress_timer.stop()
        self.bulk_update_progress.reset()

        # close the pipes
        self.bulk_update_rx_pipe.close()
        self.bulk_update_tx_pipe.close()
        self.cancel_rx_pipe.close()
        self.cancel_tx_pipe.close()
        self.cancel_rx_pipe = None
        self.cancel_tx_pipe = None

        if self.state == RadioState.SCANNING:
            self.go_to_disconnected_state()



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'setting_dialog.ui'
##
## Created by: Qt User Interface Compiler version 5.15.2
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide2.QtCore import *
from PySide2.QtGui import *
from PySide2.QtWidgets import *


class Ui_SettingDialog(object):
    def setupUi(self, SettingDialog):
        if not SettingDialog.objectName():
            SettingDialog.setObjectName(u"SettingDialog")
        SettingDialog.resize(400, 53)
        self.verticalLayout = QVBoxLayout(SettingDialog)
        self.verticalLayout.setObjectName(u"verticalLayout")
        self.tabWidget = QTabWidget(SettingDialog)
        self.tabWidget.setObjectName(u"tabWidget")

        self.verticalLayout.addWidget(self.tabWidget)

        self.buttonBox = QDialogButtonBox(SettingDialog)
        self.buttonBox.setObjectName(u"buttonBox")
        self.buttonBox.setOrientation(Qt.Horizontal)
        self.buttonBox.setStandardButtons(QDialogButtonBox.Cancel|QDialogButtonBox.Ok|QDialogButtonBox.RestoreDefaults)

        self.verticalLayout.addWidget(self.buttonBox)


        self.retranslateUi(SettingDialog)
        self.buttonBox.accepted.connect(SettingDialog.accept)
        self.buttonBox.rejected.connect(SettingDialog.reject)

        QMetaObject.connectSlotsByName(SettingDialog)
    # setupUi

    def retranslateUi(self, SettingDialog):
        SettingDialog.setWindowTitle(QCoreApplication.translate("SettingDialog", u"Device Settings", None))
    # retranslateUi




# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'change_stream_dialog.ui'
##
## Created by: Qt User Interface Compiler version 5.15.2
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide2.QtCore import *
from PySide2.QtGui import *
from PySide2.QtWidgets import *


class Ui_ChangeStreamDialog(object):
    def setupUi(self, ChangeStreamDialog):
        if not ChangeStreamDialog.objectName():
            ChangeStreamDialog.setObjectName(u"ChangeStreamDialog")
        ChangeStreamDialog.resize(403, 49)
        self.verticalLayout_2 = QVBoxLayout(ChangeStreamDialog)
        self.verticalLayout_2.setObjectName(u"verticalLayout_2")
        self.verticalLayout = QVBoxLayout()
        self.verticalLayout.setObjectName(u"verticalLayout")

        self.verticalLayout_2.addLayout(self.verticalLayout)

        self.buttonBox = QDialogButtonBox(ChangeStreamDialog)
        self.buttonBox.setObjectName(u"buttonBox")
        self.buttonBox.setOrientation(Qt.Horizontal)
        self.buttonBox.setStandardButtons(QDialogButtonBox.Cancel|QDialogButtonBox.Ok)

        self.verticalLayout_2.addWidget(self.buttonBox)


        self.retranslateUi(ChangeStreamDialog)
        self.buttonBox.accepted.connect(ChangeStreamDialog.accept)
        self.buttonBox.rejected.connect(ChangeStreamDialog.reject)

        QMetaObject.connectSlotsByName(ChangeStreamDialog)
    # setupUi

    def retranslateUi(self, ChangeStreamDialog):
        ChangeStreamDialog.setWindowTitle(QCoreApplication.translate("ChangeStreamDialog", u"Change Streams", None))
    # retranslateUi




# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'rf_power_panel.ui'
##
## Created by: Qt User Interface Compiler version 5.15.2
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide2.QtCore import *
from PySide2.QtGui import *
from PySide2.QtWidgets import *


class Ui_RFPowerPanel(object):
    def setupUi(self, RFPowerPanel):
        if not RFPowerPanel.objectName():
            RFPowerPanel.setObjectName(u"RFPowerPanel")
        RFPowerPanel.resize(111, 99)
        self.verticalLayout = QVBoxLayout(RFPowerPanel)
        self.verticalLayout.setObjectName(u"verticalLayout")
        self.enableButton = QPushButton(RFPowerPanel)
        self.enableButton.setObjectName(u"enableButton")

        self.verticalLayout.addWidget(self.enableButton)

        self.disableButton = QPushButton(RFPowerPanel)
        self.disableButton.setObjectName(u"disableButton")

        self.verticalLayout.addWidget(self.disableButton)

        self.ctrlVarLayout = QVBoxLayout()
        self.ctrlVarLayout.setObjectName(u"ctrlVarLayout")

        self.verticalLayout.addLayout(self.ctrlVarLayout)

        self.verticalSpacer = QSpacerItem(10, 10, QSizePolicy.Minimum, QSizePolicy.Expanding)

        self.verticalLayout.addItem(self.verticalSpacer)


        self.retranslateUi(RFPowerPanel)

        QMetaObject.connectSlotsByName(RFPowerPanel)
    # setupUi

    def retranslateUi(self, RFPowerPanel):
        RFPowerPanel.setWindowTitle(QCoreApplication.translate("RFPowerPanel", u"RF Power Panel", None))
        RFPowerPanel.setTitle(QCoreApplication.translate("RFPowerPanel", u"RF Power", None))
        self.enableButton.setText(QCoreApplication.translate("RFPowerPanel", u"Enable RF Power", None))
        self.disableButton.setText(QCoreApplication.translate("RFPowerPanel", u"Disable RF Power", None))
    # retranslateUi




# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'calibration_panel.ui'
##
## Created by: Qt User Interface Compiler version 5.15.2
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide2.QtCore import *
from PySide2.QtGui import *
from PySide2.QtWidgets import *


class Ui_CalibrationPanel(object):
    def setupUi(self, CalibrationPanel):
        if not CalibrationPanel.objectName():
            CalibrationPanel.setObjectName(u"CalibrationPanel")
        CalibrationPanel.resize(95, 158)
        self.verticalLayout = QVBoxLayout(CalibrationPanel)
        self.verticalLayout.setObjectName(u"verticalLayout")
        self.buttonBox = QDialogButtonBox(CalibrationPanel)
        self.buttonBox.setObjectName(u"buttonBox")
        self.buttonBox.setStandardButtons(QDialogButtonBox.Save)

        self.verticalLayout.addWidget(self.buttonBox)


        self.retranslateUi(CalibrationPanel)

        QMetaObject.connectSlotsByName(CalibrationPanel)
    # setupUi

    def retranslateUi(self, CalibrationPanel):
        CalibrationPanel.setWindowTitle(QCoreApplication.translate("CalibrationPanel", u"Calibration", None))
        CalibrationPanel.setTitle(QCoreApplication.translate("CalibrationPanel", u"Calibration", None))
    # retranslateUi

